from asyncio import sleep
from functools import wraps
from io import StringIO
from traceback import print_exc
from typing import Any, Optional, Union

from msgpack import ExtType
from nonebot.utils import logger_wrapper

from .const import ADAPTER_NAME

EXT = {0: lambda _: int.from_bytes(_, byteorder="big")}


def unpack(data: Union[ExtType, any]) -> Any:
    if isinstance(data, ExtType) and data.code in EXT:
        return EXT[data.code](data.data)
    return data


class Log:
    def __init__(self):
        self.log = logger_wrapper(ADAPTER_NAME)

    def __call__(self, level: str, message: str, exception: Optional[Exception] = None):
        self.log(level, message, exception)

    def success(self, *args, exception: Optional[Exception] = None):
        _ = StringIO()
        print(*args, file=_, end="")
        self.log("SUCCESS", _.getvalue(), exception)

    def info(self, *args, exception: Optional[Exception] = None):
        _ = StringIO()
        print(*args, file=_, end="")
        self.log("INFO", _.getvalue(), exception)

    def debug(self, *args, exception: Optional[Exception] = None):
        _ = StringIO()
        print(*args, file=_, end="")
        self.log("DEBUG", _.getvalue(), exception)

    def error(self, *args, exception: Optional[Exception] = None):
        _ = StringIO()
        print(*args, file=_, end="")
        self.log("ERROR", _.getvalue(), exception)

    def warning(self, *args, exception: Optional[Exception] = None):
        _ = StringIO()
        print(*args, file=_, end="")
        self.log("WARNING", _.getvalue(), exception)

    def trace(self, *args, exception: Optional[Exception] = None):
        _ = StringIO()
        print(*args, file=_, end="")
        self.log("TRACE", _.getvalue(), exception)


log = Log()


def retry(wait_time: int, func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            await func(*args, **kwargs)
        except Exception as e:
            log.error(f"Error when exec {func.__name__}: {repr(e)}")
            print_exc()
        await sleep(wait_time)

    return wrapper



from typing import Optional

from pydantic import BaseModel, Field, HttpUrl


class BotInfo(BaseModel):
    url: HttpUrl

    appId: Optional[str] = Field(default=None)
    appSecret: Optional[str] = Field(default=None)
    jwt: Optional[str] = Field(default=None)
    useHttp: bool = Field(default=True, description="用http<True>/socketio<False>调用接口")


class Config(BaseModel):
    bots_info: list[BotInfo] = Field(alias="tailchat_bots", default_factory=list)
    reconnect_interval: int = Field(default=5, description="重连间隔", alias="tailchat_reconnect_interval")
    time_out: int = Field(default=5, description="超时时间", alias="tailchat_time_out")



import json
from asyncio import TimeoutError as AsyncTimeoutError
from asyncio import create_task, gather, sleep, wait_for
from functools import partial
from traceback import print_exc
from typing import Any, Optional

from nonebot import escape_tag, get_plugin_config
from nonebot.adapters import Adapter as BaseAdapter
from nonebot.drivers import Driver
from nonebot.internal.driver import ContentTypes, HTTPClientMixin, Request, Response
from nonebot.message import handle_event
from yarl import URL

from .bot import Bot
from .config import BotInfo, Config
from .const import ADAPTER_NAME, Undefined
from .event import get_event
from .exception import ConnectionException, DisconnectException, get_error
from .message import Message
from .util import log, retry


class Adapter(BaseAdapter):
    def __init__(self, driver: Driver, **kwargs: Any):
        super().__init__(driver, **kwargs)
        self.adapter_config: Config = get_plugin_config(Config)
        self.tasks = []
        self.on_ready(self._setup)
        self.on_ready(Message.update_parser)
        self.driver.on_shutdown(self._shutdown)
        self.is_http_client_driver = isinstance(self.driver, HTTPClientMixin)

    @classmethod
    def get_name(cls) -> str:
        return ADAPTER_NAME

    async def _call_api(self, bot: Bot, api: str, **data: Any) -> Any:
        use_http = data.get("use_http_", bot.base_info.useHttp)  # 使用http
        use_sio = data.get("use_sio_", not use_http)  # 使用socketio
        use_api = data.get("use_api_", True)  # url / 'api' / api
        if not bot.base_info.useHttp and (not use_sio or use_http):
            log("WARNING", f"use_http is False, but api <y>{api}</y> use / strictly.")
        if use_http or not use_sio:
            if not self.is_http_client_driver:
                raise RuntimeError("HTTPClientMixin is required when use http")
            self.driver: HTTPClientMixin
            resp = await self.driver.request(
                Request(
                    "POST",
                    str((URL(bot.url) / "api" if use_api else URL(bot.url)) / api),
                    json={key: value for key, value in data.items() if value != Undefined},
                    headers={"X-Token": bot.base_info.jwt or ""},
                )
            )
            data = self._loads(resp.content)
            self._handle_http_api(resp, data)
        else:
            data = await bot.sio.call(api, {key: value for key, value in data.items() if value != Undefined})
            self._handle_socketio_api(data)
        return data.get("data")

    async def _setup(self):
        if any(i.useHttp for i in self.adapter_config.bots_info) and not self.is_http_client_driver:
            raise RuntimeError("HTTPClientMixin is required when use http")
        for i in self.adapter_config.bots_info:
            self.tasks.append(
                create_task(
                    # 防止因 socketio/tailchat 奇怪错误导致的异常
                    retry(self.adapter_config.reconnect_interval, self._handle_bot)(i),
                    name=f"handle_bot_{i.appId}",
                )
            )

    async def _shutdown(self):
        for task in self.tasks:
            if not task.done():
                task.cancel()

        await gather(*self.tasks, return_exceptions=True)

    async def _handle_bot(self, bot_info: BotInfo):
        connected = False

        async def _wait():  # 确保断连后打断wait
            nonlocal connected
            while True:
                if not connected:
                    raise DisconnectException()
                await sleep(1)

        def _check():
            nonlocal connected
            connected = False

        bot = Bot(self, bot_info.appId, bot_info)
        bot.sio.on("*", partial(self._handle_event, bot))
        bot.sio.on("disconnect", _check)
        while True:
            try:
                await wait_for(bot.login(bot.base_info.jwt), self.adapter_config.time_out)
                await wait_for(self._connect_bot(bot), self.adapter_config.time_out)
                await bot.update_info(bot.base_info.jwt)
                self._bot_connect(bot)
                connected = True
                await gather(bot.sio.wait(), _wait())
            except DisconnectException:
                log.warning(f"Bot {escape_tag(str(bot))} socketio connection closed")
            except AsyncTimeoutError:
                log.warning(f"Bot {escape_tag(str(bot))} connection timeout")
            except ConnectionException as e:
                log.error(f"Error when chat.converse.findAndJoinRoom: {repr(e)}")
            except Exception as e:
                log.error(f"Error when _handle_bot: {repr(e)}")
                print_exc()
            connected = False
            self._bot_disconnect(bot)
            log.debug(f"Bot {escape_tag(str(bot))} will reconnect in {self.adapter_config.reconnect_interval} seconds")
            await sleep(self.adapter_config.reconnect_interval)

    def _bot_connect(self, bot: Bot):
        try:
            self.bot_connect(bot)
        except Exception as e:
            log.trace(f"Error when bot_connect: {repr(e)}")

    def _bot_disconnect(self, bot: Bot):
        try:
            self.bot_disconnect(bot)
        except Exception as e:
            log.trace(f"Error when bot_disconnect: {repr(e)}")

    @staticmethod
    async def _connect_bot(bot: Bot):
        log.debug(f"try to connect bot {escape_tag(str(bot))}")
        try:
            await bot.connect()
        except Exception as e:
            log.trace(f"Error when bot.connect: {repr(e)}")
        try:
            await bot.sio.emit("chat.converse.findAndJoinRoom")
            log.success(f"<y>Bot {escape_tag(str(bot))}</y> connected")
        except Exception:
            raise ConnectionException(f"Bot {str(bot)} join room fail")

    @staticmethod
    async def _handle_event(bot: Bot, event: str, data: dict, _: Optional[any] = None):
        model = get_event(event)
        log.trace(f"Event: {event}, MatchModel: {model}, Data: {data}")
        data["event_name"] = data.get("event_name", event)
        data["self_id"] = data.get("self_id", bot.self_id)
        await handle_event(bot, await model.build(bot, data))

    @staticmethod
    def _loads(data: ContentTypes) -> Any:
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            log.error("Error when loads data", escape_tag(data.decode() if isinstance(data, bytes) else str(data)))
            raise

    @staticmethod
    def _handle_http_api(resp: Response, data: dict):
        if resp.status_code != 200:
            name = data.get("name")
            code = data.get("code")
            message = data.get("message")
            error = get_error(data.get("name"))
            name = name or "unknown"
            raise error(name=name, message=message, code=code)

    @staticmethod
    def _handle_socketio_api(data: dict):
        if not data.get("result", True):
            name = data.get("name")
            code = data.get("code")
            message = data.get("message")
            error = get_error(data.get("name"))
            name = name or "unknown"
            raise error(name=name, message=message, code=code)



from abc import ABC
from typing import Any, Literal, Union

from nonebot.adapters import Bot as BaseBot
from pydantic import TypeAdapter

from .const import Optional, Undefined
from .model import (
    Ack,
    AddFriendRequestRet,
    BaseGroupInfo,
    ClientConfig,
    ConverseInfo,
    FileInfo,
    FindAndJoinRoomRet,
    GroupAndPanelIds,
    GroupInfo,
    Health,
    InviteCodeInfo,
    LastMessages,
    MessageRet,
    Panel,
    TemporaryUserInfo,
    TokenInfo,
    UserInfo,
    Whoami,
)


class API(BaseBot, ABC):
    async def allAck(self) -> list[Ack]:
        """获取所有会话的最后一条消息的ID"""
        return TypeAdapter(list[Ack]).validate_python(await self.call_api("chat.ack.all"))

    async def allApp(self):
        """获取当前账号全部openapi app信息"""
        return await self.call_api("openapi.app.all")

    async def getApp(self, *, appId: str):
        """获取openapi app信息"""
        return await self.call_api("openapi.app.get", appId=appId)

    async def whoami(self) -> Whoami:
        """获取当前账户信息"""
        return TypeAdapter(Whoami).validate_python(await self.call_api("user.whoami"))

    async def getFile(self, *, objectName: str):
        """获取文件url (但是参数不知道填啥)"""
        return await self.call_api("file.get", objectName=objectName)

    async def ackInbox(self, *, inboxItemIds: list[str]):
        """标记消息为已读"""
        return await self.call_api("chat.inbox.ack", inboxItemIds=inboxItemIds)

    async def allInbox(self) -> MessageRet:
        """获取用户收件箱中所有内容"""
        return TypeAdapter(MessageRet).validate_python(await self.call_api("chat.inbox.all"))

    async def isMember(self, *, groupId: str):
        """是否为指定群的成员"""
        return await self.call_api("group.isMember", groupId=groupId)

    async def register(
        self,
        *,
        password: str,
        email: Optional[str] = Undefined,
        avatar: Optional[str] = Undefined,
        emailOTP: Optional[str] = Undefined,
        nickname: Optional[str] = Undefined,
        username: Optional[str] = Undefined,
    ):
        return await self.call_api(
            "user.register",
            email=email,
            avatar=avatar,
            emailOTP=emailOTP,
            nickname=nickname,
            password=password,
            username=username,
        )

    async def saveFile(self):
        return await self.call_api("file.save")

    async def statFile(self, *, objectName: str):
        return await self.call_api("file.stat", objectName=objectName)

    async def upload(self, *, file: bytes) -> FileInfo:
        """上传文件"""
        return TypeAdapter(FileInfo).validate_python(
            await self.call_api("upload", file=file, use_api_=False, use_http_=True, use_sio_=False)
        )

    async def authToken(self, *, appId: str, token: str, capability: Optional[list[str]] = Undefined):
        """验证token"""
        return await self.call_api("openapi.app.authToken", appId=appId, token=token, capability=capability)

    async def createApp(self, *, appDesc: str, appIcon: str, appName: str):
        return await self.call_api("openapi.app.create", appDesc=appDesc, appIcon=appIcon, appName=appName)

    async def deleteApp(self, *, appId: str):
        return await self.call_api("openapi.app.delete", appId=appId)

    async def loginUser(
        self, *, password: str, email: Optional[str] = Undefined, username: Optional[str] = Undefined
    ) -> TokenInfo:
        """标准登录API"""
        return TypeAdapter(TokenInfo).validate_python(
            await self.call_api("user.login", email=email, password=password, username=username)
        )

    async def quitGroup(self, *, groupId: str):
        """退群"""
        return await self.call_api("group.quitGroup", groupId=groupId)

    async def updateAck(self, *, converseId: str, lastMessageId: str):
        """更新用户在会话中已读的最后一条消息"""
        return await self.call_api("chat.ack.update", converseId=converseId, lastMessageId=lastMessageId)

    async def addBotUser(self, *, appId: str, groupId: str):
        return await self.call_api("openapi.integration.addBotUser", appId=appId, groupId=groupId)

    async def addRequest(self, *, to: str, message: Optional[str] = Undefined) -> AddFriendRequestRet:
        """发送好友申请(message好像没用, 请求处理界面看不到message).不存在的id也能发送成功.不符合格式的id会报错."""
        return TypeAdapter(AddFriendRequestRet).validate_python(
            await self.call_api("friend.request.add", to=to, message=message)
        )

    async def allRelated(self):
        return await self.call_api("friend.request.allRelated")

    async def clearInbox(self):
        """清空所有的收件箱内容"""
        return await self.call_api("chat.inbox.clear")

    async def setAppInfo(self, *, appId: str, fieldName: str, fieldValue: str):
        return await self.call_api("openapi.app.setAppInfo", appId=appId, fieldName=fieldName, fieldValue=fieldValue)

    async def getMessage(self, *, messageId: str) -> MessageRet:
        """获取消息"""
        return TypeAdapter(MessageRet).validate_python(
            await self.call_api("chat.message.getMessage", messageId=messageId)
        )

    async def addConverse(self, *, converseId: str):
        return await self.call_api("user.dmlist.addConverse", converseId=converseId)

    async def addReaction(self, *, emoji: str, messageId: str):
        """添加消息表情"""
        return await self.call_api("chat.message.addReaction", emoji=emoji, messageId=messageId)

    async def applyInvite(self, *, code: str):
        """通过邀请码加群"""
        return await self.call_api("group.invite.applyInvite", code=code)

    async def createGroup(self, *, name: str, panels: list[Union[Panel, dict]]) -> GroupInfo:
        """创建群组"""
        return TypeAdapter(GroupInfo).validate_python(
            await self.call_api("group.createGroup", name=name, panels=panels)
        )

    async def denyRequest(self, *, requestId: str):
        """拒绝好友请求"""
        return await self.call_api("friend.request.deny", requestId=requestId)

    async def getUserInfo(self, *, userId: str) -> UserInfo:
        """获取用户信息"""
        return TypeAdapter(UserInfo).validate_python(await self.call_api("user.getUserInfo", userId=userId))

    async def sendMessage(
        self,
        *,
        content: str,
        converseId: str,
        meta: Optional[dict] = Undefined,
        plain: Optional[str] = Undefined,
        groupId: Optional[str] = Undefined,
    ) -> MessageRet:
        """发送消息"""
        return TypeAdapter(MessageRet).validate_python(
            await self.call_api(
                "chat.message.sendMessage",
                meta=meta,
                plain=plain,
                content=content,
                groupId=groupId,
                converseId=converseId,
            )
        )

    async def verifyEmail(self, *, email: str):
        return await self.call_api("user.verifyEmail", email=email)

    async def configClient(self) -> ClientConfig:
        """获取客户端配置"""
        return TypeAdapter(ClientConfig).validate_python(await self.call_api("config.client"))

    async def deleteInvite(self, *, groupId: str, inviteId: str):
        """删除邀请码"""
        return await self.call_api("group.invite.deleteInvite", groupId=groupId, inviteId=inviteId)

    async def getGroupData(self, *, name: str, groupId: str):
        return await self.call_api("group.extra.getGroupData", name=name, groupId=groupId)

    async def getPanelData(self, *, name: str, groupId: str, panelId: str):
        """获取面板数据"""
        return await self.call_api("group.extra.getPanelData", name=name, groupId=groupId, panelId=panelId)

    async def isGroupOwner(self, *, groupId: str):
        """是否为创建者"""
        return await self.call_api("group.isGroupOwner", groupId=groupId)

    async def listRegistry(self):
        return await self.call_api("plugin.registry.list")

    async def removeFriend(self, *, friendUserId: str):
        """移除单项好友关系"""
        return await self.call_api("friend.removeFriend", friendUserId=friendUserId)

    async def resolveToken(self, *, token: str) -> TokenInfo:
        """获取token信息"""
        return TypeAdapter(TokenInfo).validate_python(await self.call_api("user.resolveToken", token=token))

    async def acceptRequest(self, *, requestId: str):
        """接受好友请求"""
        return await self.call_api("friend.request.accept", requestId=requestId)

    async def cancelRequest(self, *, requestId: str):
        """撤销申请好友请求. 填不存在的requestId会报错, 不符合格式的也会报错"""
        return await self.call_api("friend.request.cancel", requestId=requestId)

    async def checkIsFriend(self, *, targetId: str):
        """检查对方是否为自己好友"""
        return await self.call_api("friend.checkIsFriend", targetId=targetId)

    async def deleteMessage(self, *, messageId: str):
        """删除消息"""
        return await self.call_api("chat.message.deleteMessage", messageId=messageId)

    async def gatewayHealth(self) -> Health:
        """获取网关健康状态"""
        return TypeAdapter(Health).validate_python(await self.call_api("gateway.health"))

    async def getAllFriends(self):
        """获取全部好友的ID
        [{'id': '***'}]"""
        return await self.call_api("friend.getAllFriends")

    async def getUserGroups(self) -> list[GroupInfo]:
        """获取用户的群组"""
        return TypeAdapter(list[GroupInfo]).validate_python(await self.call_api("group.getUserGroups"))

    async def recallMessage(self, *, messageId: str) -> MessageRet:
        """撤回消息, 大于15分钟的消息无法撤回, 会报错"""
        return TypeAdapter(MessageRet).validate_python(
            await self.call_api("chat.message.recallMessage", messageId=messageId)
        )

    async def resetPassword(self, *, otp: str, email: str, password: str):
        return await self.call_api("user.resetPassword", otp=otp, email=email, password=password)

    async def saveGroupData(self, *, data: str, name: str, groupId: str):
        return await self.call_api("group.extra.saveGroupData", data=data, name=name, groupId=groupId)

    async def savePanelData(self, *, data: str, name: str, groupId: str, panelId: str):
        """保存面板数据"""
        return await self.call_api("group.extra.savePanelData", data=data, name=name, groupId=groupId, panelId=panelId)

    async def searchMessage(
        self, *, text: str, converseId: str, groupId: Optional[str] = Undefined
    ) -> list[MessageRet]:
        """搜索消息"""
        return TypeAdapter(list[MessageRet]).validate_python(
            await self.call_api("chat.message.searchMessage", text=text, groupId=groupId, converseId=converseId)
        )

    async def setAppBotInfo(self, *, appId: str, fieldName: str, fieldValue: Any):
        return await self.call_api("openapi.app.setAppBotInfo", appId=appId, fieldName=fieldName, fieldValue=fieldValue)

    async def forgetPassword(self, *, email: str):
        return await self.call_api("user.forgetPassword", email=email)

    async def getAllConverse(self) -> list[str]:
        """获取所有会话"""
        return TypeAdapter(list[str]).validate_python(await self.call_api("user.dmlist.getAllConverse"))

    async def getPermissions(self, *, groupId: str):
        return await self.call_api("group.getPermissions", groupId=groupId)

    async def modifyPassword(self, *, newPassword: str, oldPassword: str):
        """修改密码(好像没啥用)"""
        return await self.call_api("user.modifyPassword", newPassword=newPassword, oldPassword=oldPassword)

    async def removeConverse(self, *, converseId: str):
        """删除会话\n{"modifiedCount":1}"""
        return await self.call_api("user.dmlist.removeConverse", converseId=converseId)

    async def removeReaction(self, *, emoji: str, messageId: str):
        """删除消息表情"""
        return await self.call_api("chat.message.removeReaction", emoji=emoji, messageId=messageId)

    async def checkTokenValid(self, *, token: str):
        return await self.call_api("user.checkTokenValid", token=token)

    async def checkUserOnline(self, *, userIds: list[str]) -> list[bool]:
        """检查用户是否在线"""
        return TypeAdapter(list[bool]).validate_python(await self.call_api("gateway.checkUserOnline", userIds=userIds))

    async def createGroupRole(self, *, groupId: str, roleName: str, permissions: list[str]) -> GroupInfo:
        """创建群用户组"""
        return TypeAdapter(GroupInfo).validate_python(
            await self.call_api("group.createGroupRole", groupId=groupId, roleName=roleName, permissions=permissions)
        )

    async def deleteGroupRole(self, *, roleId: str, groupId: str) -> GroupInfo:
        """删除群用户组"""
        return TypeAdapter(GroupInfo).validate_python(
            await self.call_api("group.deleteGroupRole", roleId=roleId, groupId=groupId)
        )

    async def editGroupInvite(
        self, *, code: str, groupId: str, expireAt: Optional[int] = Undefined, usageLimit: Optional[int] = Undefined
    ):
        """编辑群组邀请码"""
        return await self.call_api(
            "group.invite.editGroupInvite", code=code, groupId=groupId, expireAt=expireAt, usageLimit=usageLimit
        )

    async def ensurePluginBot(self, *, botId: str, nickname: str, avatar: Optional[str] = Undefined):
        return await self.call_api("user.ensurePluginBot", botId=botId, avatar=avatar, nickname=nickname)

    async def findAndJoinRoom(self) -> FindAndJoinRoomRet:
        """查找用户相关的所有会话并加入房间"""
        return TypeAdapter(FindAndJoinRoomRet).validate_python(await self.call_api("chat.converse.findAndJoinRoom"))

    async def getUserInfoList(self, *, userIds: list[str]) -> list[UserInfo]:
        """获取多个用户信息"""
        return TypeAdapter(list[UserInfo]).validate_python(await self.call_api("user.getUserInfoList", userIds=userIds))

    async def getUserSettings(self):
        """获取用户设置"""
        return await self.call_api("user.getUserSettings")

    async def muteGroupMember(self, *, muteMs: int, groupId: str, memberId: str):
        """禁言群成员"""
        return await self.call_api("group.muteGroupMember", muteMs=muteMs, groupId=groupId, memberId=memberId)

    async def setAppOAuthInfo(self, *, appId: str, fieldName: str, fieldValue: Any):
        return await self.call_api(
            "openapi.app.setAppOAuthInfo", appId=appId, fieldName=fieldName, fieldValue=fieldValue
        )

    async def setUserSettings(self, *, settings: dict):
        """修改用户设置"""
        return await self.call_api("user.setUserSettings", settings=settings)

    async def updateUserExtra(self, *, fieldName: str, fieldValue: Any):
        return await self.call_api("user.updateUserExtra", fieldName=fieldName, fieldValue=fieldValue)

    async def updateUserField(self, *, fieldName: str, fieldValue: Any) -> UserInfo:
        """更改用户信息"""
        return TypeAdapter(UserInfo).validate_python(
            await self.call_api("user.updateUserField", fieldName=fieldName, fieldValue=fieldValue)
        )

    async def createDMConverse(self, *, memberIds: list[str]) -> ConverseInfo:
        """创建会话"""
        return TypeAdapter(ConverseInfo).validate_python(
            await self.call_api("chat.converse.createDMConverse", memberIds=memberIds)
        )

    async def createGroupPanel(
        self,
        *,
        name: str,
        type_: int,
        groupId: str,
        meta: Optional[dict] = Undefined,
        parentId: Optional[str] = Undefined,
        provider: Optional[str] = Undefined,
        pluginPanelName: Optional[str] = Undefined,
    ):
        """创建群组面板"""
        return await self.call_api(
            "group.createGroupPanel",
            name=name,
            meta=meta,
            type=type_,
            groupId=groupId,
            parentId=parentId,
            provider=provider,
            pluginPanelName=pluginPanelName,
        )

    async def deleteGroupPanel(self, *, groupId: str, panelId: str) -> GroupInfo:
        """删除群组面板"""
        return TypeAdapter(GroupInfo).validate_python(
            await self.call_api("group.deleteGroupPanel", groupId=groupId, panelId=panelId)
        )

    async def ensureOpenapiBot(self, *, botId: str, nickname: str, avatar: Optional[str] = Undefined):
        return await self.call_api("user.ensureOpenapiBot", botId=botId, avatar=avatar, nickname=nickname)

    async def findConverseInfo(self, *, converseId: str) -> ConverseInfo:
        """获取会话信息"""
        return TypeAdapter(ConverseInfo).validate_python(
            await self.call_api("chat.converse.findConverseInfo", converseId=converseId)
        )

    async def findInviteByCode(self, *, code: str) -> Optional[InviteCodeInfo]:
        return TypeAdapter(Optional[InviteCodeInfo]).validate_python(
            await self.call_api("group.invite.findInviteByCode", code=code)
        )

    async def findOpenapiBotId(self, *, email: str):
        return await self.call_api("user.findOpenapiBotId", email=email)

    async def modifyGroupPanel(
        self,
        *,
        name: str,
        type_: int,
        groupId: str,
        panelId: str,
        meta: Optional[dict] = Undefined,
        provider: Optional[str] = Undefined,
        permissionMap: Optional[dict] = Undefined,
        pluginPanelName: Optional[str] = Undefined,
        fallbackPermissions: Optional[list[str]] = Undefined,
    ):
        return await self.call_api(
            "group.modifyGroupPanel",
            meta=meta,
            name=name,
            type=type_,
            groupId=groupId,
            panelId=panelId,
            provider=provider,
            permissionMap=permissionMap,
            pluginPanelName=pluginPanelName,
            fallbackPermissions=fallbackPermissions,
        )

    async def setAppCapability(self, *, appId: str, capability: list[str]):
        return await self.call_api("openapi.app.setAppCapability", appId=appId, capability=capability)

    async def updateGroupField(self, *, groupId: str, fieldName: str, fieldValue: Any):
        """更新群组字段"""
        return await self.call_api(
            "group.updateGroupField", groupId=groupId, fieldName=fieldName, fieldValue=fieldValue
        )

    async def createGroupInvite(self, *, groupId: str, inviteType: Literal["normal", "permanent"]) -> InviteCodeInfo:
        """创建群组邀请码"""
        return TypeAdapter(InviteCodeInfo).validate_python(
            await self.call_api("group.invite.createGroupInvite", groupId=groupId, inviteType=inviteType)
        )

    async def deleteGroupMember(self, *, groupId: str, memberId: str):
        """删除群成员"""
        return await self.call_api("group.deleteGroupMember", groupId=groupId, memberId=memberId)

    async def getGroupBasicInfo(self, *, groupId: str) -> BaseGroupInfo:
        """获取群基本消息"""
        return TypeAdapter(BaseGroupInfo).validate_python(
            await self.call_api("group.getGroupBasicInfo", groupId=groupId)
        )

    async def setFriendNickname(self, *, nickname: str, targetId: str):
        """设置好友昵称"""
        return await self.call_api("friend.setFriendNickname", nickname=nickname, targetId=targetId)

    async def updateGroupConfig(self, *, groupId: str, configName: str, configValue: Any):
        """更新群组配置"""
        return await self.call_api(
            "group.updateGroupConfig", groupId=groupId, configName=configName, configValue=configValue
        )

    async def claimTemporaryUser(
        self,
        *,
        email: str,
        userId: str,
        password: str,
        emailOTP: Optional[str] = Undefined,
        username: Optional[str] = Undefined,
    ):
        return await self.call_api(
            "user.claimTemporaryUser",
            email=email,
            userId=userId,
            emailOTP=emailOTP,
            password=password,
            username=username,
        )

    async def fetchNearbyMessage(
        self, *, messageId: str, converseId: str, num: Optional[int] = Undefined, groupId: Optional[str] = Undefined
    ) -> list[MessageRet]:
        """获取指定消息的上下文"""
        return TypeAdapter(list[MessageRet]).validate_python(
            await self.call_api(
                "chat.message.fetchNearbyMessage", num=num, groupId=groupId, messageId=messageId, converseId=converseId
            )
        )

    async def verifyEmailWithOTP(self, *, emailOTP: str):
        """验证邮箱"""
        return await self.call_api("user.verifyEmailWithOTP", emailOTP=emailOTP)

    async def buildFriendRelation(self, *, user1: str, user2: str):
        """构建好友关系"""
        return await self.call_api("friend.buildFriendRelation", user1=user1, user2=user2)

    async def createTemporaryUser(self, *, nickname: str) -> TemporaryUserInfo:
        """创建临时用户"""
        return TypeAdapter(TemporaryUserInfo).validate_python(
            await self.call_api("user.createTemporaryUser", nickname=nickname)
        )

    async def updateGroupRoleName(self, *, roleId: str, groupId: str, roleName: str) -> GroupInfo:
        """更新群用户组名称"""
        return TypeAdapter(GroupInfo).validate_python(
            await self.call_api("group.updateGroupRoleName", roleId=roleId, groupId=groupId, roleName=roleName)
        )

    async def fetchConverseMessage(self, *, converseId: str, startId: Optional[str] = Undefined) -> list[MessageRet]:
        """获取会话消息"""
        return TypeAdapter(list[MessageRet]).validate_python(
            await self.call_api("chat.message.fetchConverseMessage", startId=startId, converseId=converseId)
        )

    async def getAllGroupInviteCode(self, *, groupId: str) -> list[InviteCodeInfo]:
        """获取所有群组邀请码"""
        return TypeAdapter(list[InviteCodeInfo]).validate_python(
            await self.call_api("group.invite.getAllGroupInviteCode", groupId=groupId)
        )

    async def appendGroupMemberRoles(self, *, groupId: str, roles: list[str], memberIds: list[str]):
        """添加群用户组成员"""
        return await self.call_api("group.appendGroupMemberRoles", roles=roles, groupId=groupId, memberIds=memberIds)

    async def removeGroupMemberRoles(self, *, groupId: str, roles: list[str], memberIds: list[str]):
        """移除群用户组成员"""
        return await self.call_api("group.removeGroupMemberRoles", roles=roles, groupId=groupId, memberIds=memberIds)

    async def appendDMConverseMembers(self, *, converseId: str, memberIds: list[str]):
        """在多人会话中添加成员"""
        return await self.call_api("chat.converse.appendDMConverseMembers", memberIds=memberIds, converseId=converseId)

    async def getGroupLobbyConverseId(self, *, groupId: str):
        """不知道啥作用, 但是返回的是converseId"""
        return await self.call_api("group.getGroupLobbyConverseId", groupId=groupId)

    async def searchUserWithUniqueName(self, *, uniqueName: str) -> Optional[UserInfo]:
        """根据唯一名搜索用户"""
        return TypeAdapter(Optional[UserInfo]).validate_python(
            await self.call_api("user.searchUserWithUniqueName", uniqueName=uniqueName)
        )

    async def fetchConverseLastMessages(self, *, converseIds: list[str]) -> list[LastMessages]:
        """获取多个会话的最后一条消息"""
        return TypeAdapter(list[LastMessages]).validate_python(
            await self.call_api("chat.message.fetchConverseLastMessages", converseIds=converseIds)
        )

    async def getJoinedGroupAndPanelIds(self) -> GroupAndPanelIds:
        """获取加入的群组和面板ID"""
        return TypeAdapter(GroupAndPanelIds).validate_python(await self.call_api("group.getJoinedGroupAndPanelIds"))

    async def updateGroupRolePermission(self, *, roleId: str, groupId: str, permissions: list[str]) -> GroupInfo:
        """更新群用户组权限"""
        return TypeAdapter(GroupInfo).validate_python(
            await self.call_api(
                "group.updateGroupRolePermission", roleId=roleId, groupId=groupId, permissions=permissions
            )
        )



from abc import ABC
from collections import UserDict, defaultdict, deque
from collections.abc import Iterable
from dataclasses import dataclass
from functools import wraps
from inspect import isclass
from typing import Optional, SupportsIndex, TypedDict, TypeVar, Union

from nonebot.adapters import (
    Message as BaseMessage,
)
from nonebot.adapters import (
    MessageSegment as BaseMessageSegment,
)
from typing_extensions import Self, overload

from .bbcode import Parser

BBCODE = TypeVar("BBCODE", bound="BBCode")
BBCODES: dict[str, type[BBCODE]] = {}
RELATION: dict[int, list[BBCODE]] = defaultdict(list)


def _register_bbcode(code: Union[int, type[BBCODE]]):
    if isinstance(code, int):

        def _(bbcode: type[BBCODE]) -> type[BBCODE]:
            bbcode.relation = code
            for tag in bbcode.tags:
                BBCODES[tag] = bbcode
            RELATION[code].append(bbcode)
            return bbcode

        return _
    for tag in code.tags:
        BBCODES[tag] = code
    _code = min(min(RELATION.keys() or [0]), 0) - 1
    code.relation = _code
    RELATION[_code].append(code)
    return code


class Box(TypedDict):
    text: str
    extra: dict[str, str]
    tags: list[type[BBCODE]]


class BBCode(UserDict, ABC):
    relation: int
    keys_: set[str] = set()
    tags: set[str]

    def __init__(self, *, box: Box):
        self.data: dict[str, str]  # 标签内的kv
        super().__init__(box["extra"])
        self.box = box
        for _ in self.tags:
            self.tag = _
        self.main: Optional[str] = None  # [tag=value]text[/tag]的value
        for _ in self.tags:
            self.main = self.main or self.get(_)

    @property
    def text(self) -> str:
        return self.box["text"]

    def set_text(self, text: str):
        self.box["text"] = text

    @property
    def head(self) -> str:
        return (
            "["
            + (
                (f"{self.tag}={self.main}" if self.main else self.tag)
                + " "
                + " ".join(f"{k}={v}" for k, v in self.items() if v and k in self.keys_ and k not in self.tags)
            ).rstrip(" ")
            + "]"
        )

    @property
    def tail(self) -> str:
        return f"[/{self.tag}]"

    @classmethod
    def tag_in(cls, msg: Union["MessageSegment", "Message"]) -> bool:
        if isinstance(msg, Message):
            for seg in msg:  # 递归
                if cls.tag_in(seg):
                    return True
            return False
        if cls == BBCode:  # 是BBCode
            return len(msg.tags) > 0
        elif isclass(cls):  # 子类
            return cls in msg.tags
        else:  # 实例化子类
            # 参数相同 + 在tags中
            return msg.extra | cls.data == msg.extra and type(cls) in msg.tags

    @classmethod
    def decode(cls, box: Box) -> str:
        return str(cls(box=box))

    def __str__(self):
        return self.head + self.text + self.tail

    def __repr__(self):
        return f"{self.tag}({self.data})" if len(self.data) else f"{self.tag}"

    def __missing__(self, key) -> None:
        return None


@_register_bbcode
class Markdown(BBCode):
    tags = {"md", "markdown"}


@_register_bbcode
class Card(BBCode):
    tags = {"card"}
    keys_ = {"type", "data"}

    @property
    def type(self) -> str:
        return self["type"]

    @property
    def card_data(self) -> str:
        return self["data"]


@_register_bbcode
class File(BBCode):
    tags = {"file"}
    keys_ = {"url"}

    @property
    def url(self) -> str:
        return self["url"]


@_register_bbcode
class Emoji(BBCode):
    tags = {"emoji"}


@_register_bbcode(0)
class B(BBCode):
    tags = {"b"}


@_register_bbcode(0)
class I(BBCode):
    tags = {"i"}


@_register_bbcode(0)
class U(BBCode):
    tags = {"u"}


@_register_bbcode(0)
class S(BBCode):
    tags = {"s"}


@_register_bbcode
class At(BBCode):
    tags = {"at"}
    keys_ = {"at"}

    @property
    def uid(self) -> str:
        return self.main


@_register_bbcode
class Url(BBCode):
    tags = {"url"}
    keys_ = {"url"}

    @property
    def url(self) -> str:
        return self["url"]


@_register_bbcode
class Img(BBCode):
    tags = {"img"}
    keys_ = {"width", "height"}

    @property
    def url(self) -> str:
        return self.text


@_register_bbcode
class Code(BBCode):
    tags = {"code"}


@dataclass
class MessageSegment(BaseMessageSegment["Message"]):
    data: Box

    @classmethod
    def get_message_class(cls) -> type["Message"]:
        return Message

    def __str__(self) -> str:
        return self.data["text"]

    def get_text(self) -> str:
        return self.data["text"]

    def set_text(self, text: str):
        self.data["text"] = text

    def is_text(self) -> bool:
        return self.type == "text"

    @classmethod
    def card(cls, *, text: str, type_: str, data: str) -> Self:
        return cls.build(text, [Card], {"type": type_, "data": data})

    @classmethod
    def at(cls, *, uid: str, nickname: Optional[str] = None) -> Self:
        return cls.build(nickname or uid, [At], {"at": uid})

    @classmethod
    def file(cls, *, name: str, url: str) -> Self:
        return cls.build(name, [File], {"url": url})

    @classmethod
    def url(cls, *, url: str, text: Optional[str] = None) -> Self:
        return cls.build(text or url, [Url], {"url": url})

    @classmethod
    def panel(cls, *, groupId: str, panelId: str, text: str) -> Self:
        # 同url 例如 [url=/main/group/<groupId>/<panelId>]#大厅[/url]
        return cls.build(text, [Url], {"url": f"/main/group/{groupId}/{panelId}"})

    @classmethod
    def code(cls, text: str) -> Self:
        return cls.build(text, [Code])

    @classmethod
    def text(cls, text: str, *, b: bool = False, i: bool = False, u: bool = False, s: bool = False) -> Self:
        _: list[type[BBCODE]] = []
        if b:
            _.append(B)
        if i:
            _.append(I)
        if u:
            _.append(U)
        if s:
            _.append(S)
        return cls.build(text, _)

    @classmethod
    def emoji(cls, text: str) -> Self:
        return cls.build(text, [Emoji])

    @classmethod
    def img(
        cls, url: str, *, width: Optional[Union[str, int]] = None, height: Optional[Union[str, int]] = None
    ) -> Self:
        return cls.build(url, [Img], {"width": width, "height": height})

    @classmethod
    def md(cls, text: str) -> Self:
        return cls.build(text, [Markdown])

    @classmethod
    def build(cls, text: str, tags: list[type[BBCODE]], extra: Optional[dict[str, str]] = None) -> Self:
        if extra is None:
            extra = {}
        if (length := len(tags)) == 0:
            type_ = "text"
        elif length == 1:
            type_ = tags[0].__name__.lower()
        else:
            type_ = "rich"

        _ = cls(
            type_,
            {"text": text, "extra": extra, "tags": []},
        )
        _.data["tags"] = tags
        return _

    def tag_relation(self) -> set[int]:
        return {_.relation for _ in self.tags}

    def merge_text(self, seg: Union[str, Self]) -> Self:
        self.data["text"] += seg if isinstance(self, str) else seg.get_text()
        return self

    def extend(self, seg: Self, strict: bool = True) -> Self:
        """
        新消息段的 text = self.text or seg.text
        :param seg: 要合并的消息段
        :param strict: 严格模式
        :return: 新消息段
        """
        if strict and (relations := self.tag_relation()) and any(_.relation not in relations for _ in seg.tags):
            raise ValueError("Tag relation not match")

        if self.type == "text":
            self.type = seg.type
        elif self.type != "rich":
            self.type = "rich"

        self.extra.update(seg.extra)
        self.tags.extend(seg.tags)
        self.set_text(self.get_text() or seg.get_text())
        return self

    def remove(self, bbcode: Union[BBCODE, type[BBCODE]]) -> Self:
        for _ in bbcode.keys_ | bbcode.tags:
            self.extra.pop(_, _)
        self.tags.remove(bbcode if isclass(bbcode) else type(bbcode))
        return self

    def down(self, bbcode: Union[BBCODE, type[BBCODE]]) -> Self:
        """把bbcode变为文本"""
        if bbcode is not isclass(bbcode):
            bbcode: BBCODE = self.get_tag(bbcode)
        if bbcode is None:
            return self
        bbcode: BBCODE
        self.data["text"] = str(bbcode)
        return self.remove(bbcode)

    @property
    def tags(self) -> list[type[BBCODE]]:
        return self.data["tags"]

    @property
    def extra(self) -> dict[str, str]:
        return self.data["extra"]

    def get_tag(self, bbcode: type[BBCode], index: int = 0) -> Optional[BBCODE]:
        """
        获取消息段中的BBCode
        :param bbcode: BBCode 或 子类
        :param index: 第几个，只有 bbcode=BBCode 时才有效
        :return:
        """
        if bbcode == BBCode:
            tags = self.tags
            if len(tags) > index:
                return tags[index](box=self.data)
            return None
        if bbcode in self.tags:
            return bbcode(box=self.data)
        return None

    def get_tags(self) -> list[BBCODE]:
        return [tag(box=self.data) for tag in self.tags]

    def decode(self) -> str:
        _ = deque([self.data["text"]])
        for tag in self.tags[::-1]:
            bbcode: BBCODE = tag(box=self.data)
            _.appendleft(bbcode.head)
            _.append(bbcode.tail)
        return "".join(_)

    def index(self, value: Union[str, type[BBCODE], BBCODE], raise_: bool = True) -> int:
        index = 0
        if isinstance(value, str):
            for tag in self.tags:
                if value in tag.keys_:
                    return index
                index += 1
        else:
            if not isclass(value):
                value = type(value)
            for tag in self.tags:
                if tag == value:
                    return index
                index += 1
        if raise_:
            raise ValueError(f"{value} not in list")
        return -1


def _update_parser(func):
    @wraps(func)
    def _(*args, **kwargs):
        res = func(*args, **kwargs)
        Message.parser = Parser(MessageSegment, BBCODES)
        return res

    return _


class Message(BaseMessage[MessageSegment]):
    parser: Parser = Parser(MessageSegment, BBCODES)

    @classmethod
    def get_segment_class(cls) -> type[MessageSegment]:
        return MessageSegment

    def __init__(
        self,
        message: Union[str, None, Iterable[MessageSegment], MessageSegment] = None,
    ):
        super().__init__(message)
        self.reduce()  # 合并连续的纯文本

    @classmethod
    def _construct(cls, msg: str) -> Iterable[MessageSegment]:
        yield from cls.parser.tokenize(msg)

    @classmethod
    def update_parser(cls):
        cls.parser = Parser(MessageSegment, BBCODES)

    @classmethod
    def register_bbcode(cls, code: Union[int, type[BBCODE]]):
        _register_bbcode(code)
        cls.update_parser()

    @overload
    def __getitem__(self, args: str) -> Self: ...

    @overload
    def __getitem__(self, args: tuple[str, int]) -> MessageSegment: ...

    @overload
    def __getitem__(self, args: tuple[str, slice]) -> Self: ...

    @overload
    def __getitem__(self, args: int) -> MessageSegment: ...

    @overload
    def __getitem__(self, args: slice) -> Self: ...

    @overload
    def __getitem__(self, args: type[BBCODE]) -> list[BBCODE]: ...

    @overload
    def __getitem__(self, args: tuple[type[BBCODE], int]) -> BBCODE: ...

    @overload
    def __getitem__(self, args: tuple[type[BBCODE], slice]) -> list[BBCODE]: ...

    def __getitem__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        args: Union[str, tuple[Union[str, type[BBCODE]], Union[int, slice]], int, slice, type[BBCODE]],
    ) -> Union[MessageSegment, Self, BBCODE, list[BBCODE]]:
        # 判断是否为 type[B]
        arg1 = args
        arg2 = slice(None)
        if isinstance(args, tuple):
            arg1, arg2 = args

        if isclass(arg1) and issubclass(arg1, BBCode):  # 是类 + 是BBCode
            return self.get_tag(arg1, arg2, strict=True) if isinstance(arg2, int) else self.get_tags(arg1)[arg2]
        return super().__getitem__(args)

    def has(self, value: Union[MessageSegment, str, type[BBCODE], BBCODE]) -> bool:
        return value in self

    def __contains__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, value: Union[MessageSegment, str, type[BBCODE], BBCODE]
    ) -> bool:
        if isinstance(value, (MessageSegment, str)):
            return super().__contains__(value)
        return value.tag_in(self)

    def index(self, value: Union[MessageSegment, str, type[BBCODE], BBCODE], *args: SupportsIndex) -> int:
        if isinstance(value, (MessageSegment, str)):
            return super().index(value, *args)
        index = 0
        for seg in self:
            _index = seg.index(value, raise_=False)
            if _index != -1:
                return _index + index
            index += len(seg.tags)
        raise ValueError(f"{value} not in list")

    def count(self, value: Union[MessageSegment, str, type[BBCode]]) -> int:
        if isinstance(value, (MessageSegment, str)):
            return super().count(value)
        return sum(1 for seg in self if value in seg.tags)

    def only(self, value: Union[MessageSegment, str, type[BBCODE]]) -> bool:
        if isinstance(value, (MessageSegment, str)):
            return super().__contains__(value)
        for seg in self:
            for tag in seg.tags:
                if tag != value:
                    return False
        return True

    def reduce(self):
        # MIT https://github.com/nonebot/adapter-onebot/blob/v2.4.3/nonebot/adapters/onebot/v11/message.py#L334-L342
        """合并消息内连续的纯文本段。"""
        index = 1
        while index < len(self):
            if len(self[index - 1].tags) == 0 and len(self[index].tags) == 0:  # 连续的纯文本
                self[index - 1].merge_text(self[index])
                del self[index]
            else:
                index += 1

    def merge_text(self) -> Self:
        if len(self) < 2:  # 只有一个消息段
            return self
        _ = [self[0]]
        for seg in self[1:]:
            seg: MessageSegment
            if len(seg.tags) == 0 and len(_[-1].tags) == 0:  # 连续的纯文本
                _[-1].merge_text(seg)
                continue
            _.append(seg)
        return Message(_)

    def extract_plain_text(self) -> str:
        return "".join(seg.data["text"] for seg in self if seg.is_text())

    def decode(self) -> str:
        return "".join(seg.decode() for seg in self)

    def show(self) -> str:
        return " ".join(seg.decode() for seg in self)

    def get_tag(self, bbcode: type[BBCode], index: int = 0, strict: bool = False) -> Optional[BBCODE]:
        if bbcode == BBCode:
            for seg in self:
                for tag in seg.tags:
                    if index == 0:
                        return tag(box=seg.data)
                    index -= 1
            if strict:
                raise IndexError(f"{bbcode} not in list")
            return None
        for seg in self:
            for tag in seg.tags:
                if tag == bbcode:
                    if index == 0:
                        return bbcode(box=seg.data)
                    index -= 1
        if strict:
            raise IndexError(f"{bbcode} not in list")
        return None

    def get_tags(self, bbcode: type[BBCode]) -> list[BBCODE]:
        if bbcode == BBCode:
            return [tag(box=seg.data) for seg in self for tag in seg.tags]
        return [tag(box=seg.data) for seg in self for tag in seg.tags if tag == bbcode]

    @property
    def tags(self) -> list[type[BBCODE]]:
        return [tag for seg in self for tag in seg.tags]



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal, Optional, TypeVar, Union, cast, get_args

from nonebot.adapters import Event as BaseEvent
from nonebot.compat import model_dump
from pydantic import Field
from pydantic_core import PydanticUndefined
from typing_extensions import Self

from .message import At, Message, MessageSegment
from .model import (
    Announcement,
    ConverseType,
    GroupInfo,
    MessageMeta,
    MessageRet,
    ObjectId,
    Payload,
    Reaction,
    Replay,
    StaticAnnouncement,
    datetime,
)
from .util import log

if TYPE_CHECKING:
    from .bot import Bot


class Event(BaseEvent, ABC):
    event_name: str
    event_type: str

    self_id: ObjectId

    _isToMe: bool = False

    def get_type(self) -> str:
        return self.event_type

    def get_event_name(self) -> str:
        return self.event_name

    def get_event_description(self) -> str:
        return str(model_dump(self))

    def is_tome(self) -> bool:
        return not self.is_self() and (self._isToMe or self._is_tome)

    @property
    @abstractmethod
    def _is_tome(self) -> bool:
        return True

    def is_self(self) -> bool:
        return self.self_id == self.get_user_id()

    async def _get_message(self, bot: "Bot" = None) -> Message:
        return self.get_message()

    @classmethod
    async def build(cls, bot: "Bot", obj: dict) -> Self:
        event = cls.model_validate(obj)
        try:
            message: Message = await event._get_message(bot)
            # at
            seg: MessageSegment = message[0]
            if seg.type == At.__name__.lower():
                if seg.get_tag(At).main == event.self_id:
                    event._isToMe = True
                    message.pop(0)
            # nickname
            elif seg.type == "text":
                text = seg.get_text().strip()
                for i in bot.config.nickname:
                    if text.startswith(i):
                        event._isToMe = True
                        if len(text) == len(i):
                            message.pop(0)
                        else:
                            seg.data["text"] = text[len(i) :]
        except Exception as e:
            log.debug(f"Event.build error: {e}")
        return event


class NoticeEvent(Event, ABC):
    event_type: Literal["notice"] = "notice"

    def get_message(self) -> Message:
        raise ValueError("Event has no message!")

    def get_session_id(self) -> str:
        raise ValueError("Event has no context!")


class RequestEvent(Event, ABC):
    event_type: Literal["request"] = "request"

    def get_session_id(self) -> str:
        raise ValueError("Event has no context!")

    def get_message(self) -> Message:
        raise ValueError("Event has no context!")


class MessageEvent(Event, ABC):
    event_type: Literal["message"] = "message"

    def get_plaintext(self) -> str:
        return self.get_message().extract_plain_text()

    @abstractmethod
    def get_message_id(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_converse_id(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def is_group(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_group_id(self) -> Optional[str]:
        raise NotImplementedError

    def get_session_id(self) -> str:
        return f"{self.get_group_id()}:{self.get_converse_id()}" if self.is_group() else self.get_converse_id()


class UnknownEvent(Event):
    event_type: Literal["unknown"] = "unknown"

    @property
    def _is_tome(self) -> bool:
        return False

    def get_user_id(self) -> str:
        raise ValueError("Unknown event has no context!")

    def get_session_id(self) -> str:
        raise ValueError("Unknown event has no context!")

    def get_message(self) -> "Message":
        if hasattr(self, "content"):
            if not isinstance(self.content, Message):
                setattr(self, "content", Message(self.content))
            return self.content
        raise ValueError("Unknown event has no message content")


EVENT_CLASSES: dict[str, type[Event]] = {}

E = TypeVar("E", bound=Event)


def get_event(event_name: str) -> type[Event]:
    return EVENT_CLASSES.get(event_name, UnknownEvent)


def register_event_class(event_class: type[E]) -> type[E]:
    if event_class.model_fields["event_type"].default == PydanticUndefined:
        raise ValueError(f"Event class {event_class} must have a field `event_type`")
    try:
        event_name = getattr(event_class, "event_name", get_args(event_class.__annotations__["event_name"])[0])
    except Exception as e:
        raise ValueError(
            f"Event class {event_class} must have a default value or Literal[str] type annotation for `event_name`"
        ) from e  # event_name 必须 有默认值 或 Literal[str] 类型注解
    EVENT_CLASSES[event_name] = event_class
    return event_class


@register_event_class
class ClientConfigUpdateEvent(NoticeEvent):
    event_name: Literal["notify:config.updateClientConfig"]

    serverName: str
    serverEntryImage: str
    announcement: Union[StaticAnnouncement, Literal[False]]

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        return "000000000000000000000000"


@register_event_class
class DMConverseUpdateEvent(NoticeEvent):
    """创建私信会话时的事件"""

    event_name: Literal["notify:chat.converse.updateDMConverse"]

    id: ObjectId = Field(alias="_id")  # 应该是converseId
    type: ConverseType  # DM, Multi
    members: list[ObjectId]
    createdAt: datetime
    updatedAt: datetime

    v: int = Field(alias="__v")

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        raise ValueError("Event has no context!")


@register_event_class
class MessageDeleteEvent(NoticeEvent):
    event_name: Literal["notify:chat.message.delete"]

    converseId: ObjectId
    messageId: ObjectId

    @property
    def _is_tome(self) -> bool:
        # TODO: 提供信息过少
        return False

    def get_user_id(self) -> str:
        # TODO: 提供信息过少
        raise ValueError("Event has no context!")


class GroupInfoEvent(NoticeEvent, ABC):
    @property
    def _is_tome(self) -> bool:
        # TODO: 不好判断
        return False


@register_event_class
class AddGroupEvent(GroupInfoEvent, GroupInfo):
    """机器人加群事件"""

    event_name: Literal["notify:group.add"]

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        return self.self_id


@register_event_class
class RemoveGroupEvent(GroupInfoEvent):
    """机器人退群事件"""

    event_name: Literal["notify:group.remove"]

    groupId: ObjectId

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        return self.self_id


@register_event_class
class GroupInfoUpdateEvent(GroupInfoEvent, GroupInfo):
    """群信息更新/用户加群事件"""

    event_name: Literal["notify:group.updateInfo"]

    def get_user_id(self) -> str:
        # TODO: ...
        raise ValueError("Event has no context!")


class DefaultReactionEvent(NoticeEvent):
    converseId: ObjectId
    messageId: ObjectId
    reaction: Reaction

    message: Optional[MessageRet] = None
    content: Optional[Message] = None

    def get_user_id(self) -> str:
        return self.reaction.author

    @property
    def _is_tome(self) -> bool:
        if self.message:
            return self.message.author == self.self_id
        return False

    def get_session_id(self) -> str:
        return f"{self.converseId}:{self.messageId}:{self.reaction.author}"

    def is_group(self) -> bool:
        if self.message is None:
            raise ValueError("Event has no context!")
        return not not self.message.groupId

    def get_message(self) -> Message:
        if self.content:
            return self.content
        return super().get_message()

    async def _get_message(self, bot: Optional["Bot"] = None) -> Message:
        if self.content:
            return self.content
        if self.message:
            self.content = Message(self.message.content)
            return self.content
        if bot:
            self.message = await bot.getMessage(messageId=self.messageId)
            self.content = Message(self.message.content)
            return self.content
        raise ValueError("Event has no context!")


@register_event_class
class DefaultReactionAddEvent(DefaultReactionEvent):
    event_name: Literal["notify:chat.message.addReaction"]

    @classmethod
    async def build(cls, bot: "Bot", obj: dict) -> Self:
        event = cast(cls, await super().build(bot, obj))
        if event.message is None:
            return event
        if event.is_group():
            return GroupReactionAddEvent.model_validate(dict(event))
        return PrivateReactionAddEvent.model_validate(dict(event))


class PrivateReactionAddEvent(DefaultReactionAddEvent):
    message: MessageRet
    content: Message

    def is_group(self) -> bool:
        return False


class GroupReactionAddEvent(DefaultReactionAddEvent):
    message: MessageRet
    content: Message

    def is_group(self) -> bool:
        return True

    @property
    def groupId(self) -> ObjectId:
        return self.message.groupId


@register_event_class
class DefaultReactionRemoveEvent(DefaultReactionEvent):
    event_name: Literal["notify:chat.message.removeReaction"]

    @classmethod
    async def build(cls, bot: "Bot", obj: dict) -> Self:
        event = cast(cls, await super().build(bot, obj))
        if event.message is None:
            return event
        if event.is_group():
            return GroupReactionRemoveEvent.model_validate(dict(event))
        return PrivateReactionRemoveEvent.model_validate(dict(event))


class PrivateReactionRemoveEvent(DefaultReactionRemoveEvent):
    message: MessageRet
    content: Message

    def is_group(self) -> bool:
        return False


class GroupReactionRemoveEvent(DefaultReactionRemoveEvent):
    message: MessageRet
    content: Message

    def is_group(self) -> bool:
        return True

    @property
    def groupId(self) -> ObjectId:
        return self.message.groupId


class DefaultMessageEvent(MessageEvent):
    id: ObjectId = Field(alias="_id")
    author: ObjectId
    hasRecall: bool
    converseId: ObjectId
    meta: Optional[MessageMeta] = Field(default=None)

    content: Message
    raw_content: str = Field(alias="content")
    reactions: list[Reaction]
    createdAt: datetime
    updatedAt: datetime

    groupId: Optional[ObjectId] = Field(default=None)

    def get_message_id(self) -> str:
        return self.id

    def get_converse_id(self) -> str:
        return self.converseId

    def get_group_id(self) -> Optional[str]:
        return self.groupId

    @property
    def replay(self) -> Optional[Replay]:
        return self.meta.replay if self.meta else None

    def get_message(self) -> Message:
        return self.content

    def get_user_id(self) -> str:
        return self.author

    def _is_tome(self) -> bool:
        # 私聊/提及/回复
        return (
            not self.is_group()
            or (self.meta and self.self_id in self.meta.mentions)
            or (self.meta and self.meta.replay and self.self_id == self.meta.replay.author)
        )

    def get_event_description(self) -> str:
        return (
            f"Message {self.id} from {self.author}"
            + (f"@[群:{self.groupId}]" if self.groupId else "")
            + f" {self.content.show()}"
        )

    def is_group(self) -> bool:
        return not not self.groupId


@register_event_class
class MessageAddEvent(DefaultMessageEvent):
    event_name: Literal["notify:chat.message.add"]


@register_event_class
class MessageUpdateEvent(DefaultMessageEvent):
    """消息更新事件(撤回消息)"""

    event_name: Literal["notify:chat.message.update"]


class AtEvent(Event, ABC):
    event_type: Literal["message"] = "message"

    type: str

    v: int = Field(alias="__v")


@register_event_class
class AtMessageEvent(AtEvent, MessageEvent):
    event_name: Literal["notify:chat.inbox.append"]

    id: ObjectId = Field(alias="_id")
    userId: ObjectId
    payload: Union[Payload, Announcement]
    readed: bool
    createdAt: datetime
    updatedAt: datetime

    def get_message(self) -> Message:
        return self.payload.messageSnippet

    def get_message_id(self) -> str:
        return self.payload.messageId

    def get_converse_id(self) -> str:
        return self.payload.converseId

    def get_group_id(self) -> Optional[str]:
        return self.payload.groupId

    def is_group(self):
        return not not self.payload.groupId

    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        return self.payload.messageAuthor

    def get_event_description(self) -> str:
        return (
            f"AtMessage {self.payload.messageId} from {self.payload.messageAuthor}"
            + (f"@[群:{self.get_group_id()}]" if self.payload.groupId else "")
            + f" {self.payload.messageSnippet.show()}"
        )


@register_event_class
class AtMessageUpdateEvent(AtEvent, NoticeEvent):
    """At 消息更新事件"""

    event_name: Literal["notify:chat.inbox.updated"]

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        raise ValueError("Event has no context!")


@register_event_class
class FriendRequestAddEvent(RequestEvent):
    event_name: Literal["notify:friend.request.add"]

    id: ObjectId = Field(alias="_id")
    from_: ObjectId = Field(alias="from")
    to: ObjectId
    v: int = Field(alias="__v")

    async def accept(self, bot: "Bot"):
        return await bot.acceptRequest(requestId=self.id)

    async def deny(self, bot: "Bot"):
        return await bot.denyRequest(requestId=self.id)

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        return self.from_


@register_event_class
class FriendRequestRemoveEvent(RequestEvent):
    event_name: Literal["notify:friend.request.remove"]

    requestId: ObjectId

    @property
    def _is_tome(self) -> bool:
        return True

    def get_user_id(self) -> str:
        # TODO: 提供信息过少
        raise ValueError("Event has no context!")



from base64 import urlsafe_b64decode
from functools import wraps
from hashlib import md5
from json import loads
from time import time
from typing import TYPE_CHECKING, Optional, Union

from pydantic import TypeAdapter
from socketio import AsyncClient as AsyncSocketClient

from .api import API
from .config import BotInfo
from .event import Event, MessageEvent
from .message import Message, MessageSegment
from .model import (
    BaseBotInfo,
    JwtPayload,
    MessageRet,
    TokenInfo,
)

if TYPE_CHECKING:
    from .adapter import Adapter


def _with_update_info(func):
    @wraps(func)
    async def wrapper(self: "Bot", *args, **kwargs):
        _ = await func(self, *args, **kwargs)
        await self.update_info(self.base_info.jwt)
        return _

    return wrapper


class Bot(API):
    def __str__(self):
        return f"{self.info.nickname}-{self.self_id}" if self.info else self.self_id

    async def connect(self):
        return await self.sio.connect(
            url=self.url,
            auth={"token": self.base_info.jwt},
            transports=["websocket"],
        )

    @staticmethod
    def md5(text: str) -> str:
        return md5(text.encode()).hexdigest()

    @staticmethod
    def decode_jwt(jwt: str) -> tuple[dict, JwtPayload, str]:
        segments = jwt.split(".")
        if len(segments) != 3:
            raise ValueError("Invalid JWT format.")

        return (
            loads(urlsafe_b64decode(segments[0] + "===").decode("utf-8")),
            JwtPayload.model_validate(loads(urlsafe_b64decode(segments[1] + "===").decode("utf-8"))),
            segments[2],
        )

    @classmethod
    async def build(cls, adapter: "Adapter", base_info: BotInfo):
        bot = Bot(adapter, base_info.appId, base_info)
        await bot.login(base_info.jwt)
        return bot

    async def update_info(self, jwt: Optional[str]):
        jwt = jwt or self.base_info.jwt
        self.info = await self.resolveToken(token=jwt)
        self.base_info.jwt = jwt
        self.self_id = self.info.userId

    def __init__(self, adapter: "Adapter", self_id: str, base_info: BotInfo):
        super().__init__(adapter, self_id)
        self.url = str(base_info.url)
        self.base_info = base_info
        self.sio = AsyncSocketClient(serializer="msgpack")
        self.info: Optional[TokenInfo] = None

    async def send(
        self,
        event: Event,
        message: Union[str, Message, MessageSegment],
        encode: bool = False,
        **kwargs,
    ) -> MessageRet:
        if not isinstance(event, MessageEvent):
            raise ValueError("event must be MessageEvent")
        return await self.sendMessage(
            content=(Message(message).decode() if encode else message)
            if isinstance(message, str)
            else message.decode(),
            converseId=event.get_converse_id(),
            groupId=event.get_group_id(),
            meta=kwargs.get("meta"),
            plain=kwargs.get("plain"),
        )

    async def loginBot(self, *, appId: str, appSecret: str) -> BaseBotInfo:
        return TypeAdapter(BaseBotInfo).validate_python(
            await self.call_api(
                "openapi.bot.login", token=self.md5(appId + appSecret), appId=appId, use_http_=True, use_sio_=False
            )
        )

    async def login(self, jwt: Optional[str] = None):
        if jwt:
            jwt_data = self.decode_jwt(jwt)[1]
            if jwt_data.exp.timestamp() > time():
                self.base_info.jwt = jwt
                return
        if self.base_info.appId and self.base_info.appSecret:
            data = await self.loginBot(appId=self.base_info.appId, appSecret=self.base_info.appSecret)
            self.base_info.jwt = data.jwt
            return
        raise ValueError("必须提供jwt或appId和appSecret")

    @_with_update_info
    async def updateNickname(self, nickname: str):
        """更新昵称，机器人重登后就失效了"""
        return await self.updateUserField(fieldName="nickname", fieldValue=nickname)

    @_with_update_info
    async def updateAvatar(self, avatar: str):
        """更新头像，机器人重登后就失效了"""
        return await self.updateUserField(fieldName="avatar", fieldValue=avatar)



from nonebot.rule import Rule

from .event import DefaultReactionEvent, Event


async def _not_self(event: Event) -> bool:
    return not event.is_self()


"""不是机器人自己的事件"""
NotSelf: Rule = Rule(_not_self)


def targetMessageReactionEvent(messageId: str) -> Rule:
    """指定消息的 ReactionEvent"""

    async def _target_message(event: DefaultReactionEvent) -> bool:
        return event.messageId == messageId

    return Rule(_target_message)



from typing import Optional

from nonebot.exception import ActionFailed as BaseActionFailed
from nonebot.exception import AdapterException

from .const import ADAPTER_NAME

ERRORS: dict[str, type["ActionFailed"]] = {}


def get_error(name: str) -> Optional[type["ActionFailed"]]:
    return ERRORS.get(name, ActionFailed)


def register_error(error: type["ActionFailed"]) -> None:
    ERRORS[error.__name__] = error


class TailchatAdapterException(AdapterException):
    def __init__(self):
        super().__init__(ADAPTER_NAME)


class ActionFailed(BaseActionFailed, TailchatAdapterException):
    def __init__(self, *, name: Optional[str], code: Optional[int] = None, message: Optional[str] = None):
        super(TailchatAdapterException).__init__()
        self.name = name
        self.code = code
        self.message = message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, code={self.code!r}, message={self.message!r})"


class DisconnectException(TailchatAdapterException): ...


class ConnectionException(TailchatAdapterException):
    def __init__(self, message: str):
        self.message = message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(message={self.message!r})"


class Error(ActionFailed):
    """通用错误"""


class DataNotFoundError(ActionFailed):
    """找不到~"""


class CastError(ActionFailed):
    """填的参数不符合规范"""


for _ in ActionFailed.__subclasses__():
    ERRORS[_.__name__] = _



from datetime import datetime as raw_datetime
from typing import Annotated, Literal, Optional

from pydantic import BaseModel, BeforeValidator, ByteSize, ConfigDict, Field, RootModel

from .config import BotInfo
from .message import Message, MessageSegment
from .util import unpack

datetime = Annotated[
    raw_datetime,
    BeforeValidator(unpack),
]

ObjectId = Annotated[str, Field(min_length=24, max_length=24)]
ConverseType = Literal[
    "DM",  # 私信
    "Multi",  # 多人
    "Group",  # 群组
]


class RawModel(BaseModel):
    model_config = ConfigDict(extra="allow")


class MemberInfo(RawModel):
    roles: list[str]
    userId: str

    muteUntil: Optional[datetime] = None


class Panel(RawModel):
    id: str
    name: str
    type: int
    fallbackPermissions: list[str]

    parentId: Optional[str] = None
    pluginPanelName: Optional[str] = None
    provider: Optional[str] = None


class Replay(RawModel):
    id: str = Field(alias="_id")
    content: str
    author: str


class Emoji(RootModel[str]):
    @property
    def name(self):
        return self.root[1:-1]

    def to_seg(self) -> MessageSegment:
        return MessageSegment.emoji(self.name)

    def __str__(self):
        return self.root


class Reaction(RawModel):
    name: Emoji
    author: str
    id: Optional[str] = Field(default=None, alias="_id")


class MessageMeta(RawModel):
    mentions: list[str]

    replay: Optional[Replay] = None


class Payload(RawModel):
    converseId: str
    messageId: str
    messageAuthor: str
    messageSnippet: Message
    messagePlainContent: Message

    groupId: Optional[str] = None


class Announcement(RawModel):
    title: str
    content: str  # md富文本


class StaticAnnouncement(RawModel):
    id: datetime
    text: str
    link: str


class ClientInfo(BotInfo):
    id: str = Field(alias="_id")
    login_time: datetime
    expire_time: datetime
    email: str
    userId: str
    nickname: str
    avatar: str


class BaseUserInfo(RawModel):
    userId: str = Field(alias="_id")
    nickname: str
    email: str

    avatar: Optional[str] = Field(default=None)


class BaseBotInfo(BaseUserInfo):
    jwt: str
    userId: str
    email: str
    nickname: str
    avatar: str  # bot avatar 默认为''


class JwtPayload(BaseUserInfo):
    avatar: str  # bot avatar 默认为''
    iat: datetime
    exp: datetime


class TokenInfo(RawModel):
    """token=jwt"""

    userId: str = Field(alias="_id")
    email: str
    nickname: str
    avatar: str
    temporary: bool
    type: str
    emailVerified: bool
    banned: bool
    createdAt: datetime
    token: str

    discriminator: Optional[str] = None


TemporaryUserInfo = TokenInfo


class MessageRet(RawModel):
    id: str = Field(alias="_id")
    content: str
    author: str
    converseId: str
    hasRecall: bool
    reactions: list[Reaction]
    createdAt: datetime
    updatedAt: datetime

    v: int = Field(alias="__v")

    groupId: Optional[str] = None
    meta: Optional[MessageMeta] = None


class ClientConfig(RawModel):
    uploadFileLimit: ByteSize
    emailVerification: bool
    disableUserRegister: bool
    disableGuestLogin: bool
    disableCreateGroup: bool
    disablePluginStore: bool
    disableAddFriend: bool
    disableTelemetry: bool
    serverName: str


class Ack(RawModel):
    userId: str
    converseId: str
    lastMessageId: str


class UserInfo(BaseUserInfo):
    temporary: bool  # 临时用户
    type: str
    emailVerified: bool
    banned: bool
    createdAt: datetime

    discriminator: Optional[str] = None  # eya46#1145 -> nickname#discriminator


class Whoami(RawModel):
    userAgent: str
    language: str
    user: BaseUserInfo
    token: str
    userId: str


class ConverseInfo(RawModel):
    id: str = Field(alias="_id")
    type: str
    members: list[str]
    createdAt: datetime
    updatedAt: datetime

    v: int = Field(alias="__v")

    name: Optional[str]


class LastMessages(RawModel):
    converseId: str
    lastMessageId: str


class Cpu(RawModel):
    load1: float
    load5: float
    load15: float
    cores: int
    utilization: int


class Memory(RawModel):
    free: ByteSize
    total: ByteSize
    percent: float


class ProcessMemory(RawModel):
    rss: ByteSize
    heapTotal: ByteSize
    heapUsed: ByteSize
    external: ByteSize
    arrayBuffers: ByteSize


class Process(RawModel):
    pid: int
    memory: ProcessMemory
    uptime: float
    argv: list[str]


class Health(RawModel):
    nodeID: str
    instanceID: str  # uuid
    cpu: Cpu
    memory: Memory
    process: Process
    services: list[str]
    version: str


class FileInfo(RawModel):
    etag: str
    path: str
    url: str


class InviteCodeInfo(RawModel):
    id: str = Field(alias="_id")
    code: str
    creator: str
    groupId: str
    usage: int
    createdAt: datetime
    updatedAt: datetime

    v: int = Field(alias="__v")


class BaseGroupInfo(RawModel):
    name: str
    owner: str
    description: str
    memberCount: int


class Role(RawModel):
    id: str = Field(alias="_id")
    name: str
    permissions: list[str]


class GroupInfo(RawModel):
    id: str = Field(alias="_id")
    name: str
    owner: str
    members: list[MemberInfo]
    panels: list[Panel]
    roles: list[Role]
    fallbackPermissions: list[str]
    createdAt: datetime
    updatedAt: datetime

    v: int = Field(alias="__v")

    config: Optional[dict] = None
    description: Optional[str] = None
    avatar: Optional[str] = None


class AddFriendRequestRet(RawModel):
    id: str = Field(alias="_id")  # requestId
    from_: str = Field(alias="from")
    to: str = Field(alias="to")
    v: int = Field(alias="__v")


class GroupAndPanelIds(RawModel):
    groupIds: list[str]
    textPanelIds: list[str]
    subscribeFeaturePanelIds: list[str]


class FindAndJoinRoomRet(RawModel):
    dmConverseIds: list[ObjectId]
    groupIds: list[ObjectId]
    textPanelIds: list[ObjectId]
    subscribeFeaturePanelIds: list[ObjectId]



# BSD-2-Clause license
# https://github.com/dcwatson/bbcode/blob/master/bbcode.py
"""
Copyright (c) 2011, Dan Watson.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from collections import OrderedDict, defaultdict, deque
from shlex import split
from sys import getrecursionlimit
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .message import BBCODE, MessageSegment


class Parser:
    def __init__(
        self,
        seg: type["MessageSegment"],
        tags: dict[str, type["BBCODE"]],
        drop_unrecognized=False,
        max_tag_depth=None,
    ):
        self.seg = seg
        self.tag_opener = "["
        self.tag_closer = "]"
        self.tags = tags
        self.drop_unrecognized = drop_unrecognized
        self.max_tag_depth = max_tag_depth or getrecursionlimit()

    def _parse_opts(self, data):
        opts = OrderedDict()
        args = split(data)
        tag, args = args[0], args[1:]

        if "=" in tag:
            name, opts[name.lower()] = tag.split("=", 1)
        else:
            name = tag

        for arg in args:
            if "=" in arg:
                k, v = arg.split("=", 1)
                opts[k.lower()] = v
            else:
                opts[arg.lower()] = ""
        return name.lower(), opts

    def _parse_tag(self, tag):
        """
        Given a tag string (characters enclosed by []), this function will
        parse any options and return a tuple of the form:
            (valid, tag_name, closer, options)
        """
        if not tag.startswith(self.tag_opener) or not tag.endswith(self.tag_closer):
            return False, tag, False, None
        tag_name = tag[len(self.tag_opener) : -len(self.tag_closer)].strip()
        if not tag_name:
            return False, tag, False, None
        closer = False
        opts = {}
        if tag_name[0] == "/":
            tag_name = tag_name[1:]
            closer = True
        # Parse options inside the opening tag, if needed.
        if (("=" in tag_name) or (" " in tag_name)) and not closer:
            tag_name, opts = self._parse_opts(tag_name)
        return True, tag_name.strip().lower(), closer, opts

    def _tag_extent(self, data, start):
        """
        Finds the extent of a tag, accounting for option quoting and new tags starting before the current one closes.
        Returns (found_close, end_pos) where valid is False if another tag started before this one closed.
        """
        in_quote = False
        quotable = False
        lto = len(self.tag_opener)
        ltc = len(self.tag_closer)
        for i in range(start + 1, len(data)):
            ch = data[i]
            if ch == "=":
                quotable = True
            if ch in ('"', "'"):
                if quotable and not in_quote:
                    in_quote = ch
                elif in_quote == ch:
                    in_quote = False
                    quotable = False
            if not in_quote and data[i : i + lto] == self.tag_opener:
                return i, False
            if not in_quote and data[i : i + ltc] == self.tag_closer:
                return i + ltc, True
        return len(data), False

    def tokenize(self, data):
        data = data.replace("\r\n", "\n").replace("\r", "\n")
        pos = 0
        ld = len(data)
        tokens: list[Union["MessageSegment", tuple["MessageSegment", str, str]]] = []
        tags: dict[str, deque["MessageSegment"]] = defaultdict(deque)
        while pos < ld:
            start = data.find(self.tag_opener, pos)  # 寻找下一个标签开始

            if start >= pos:
                if start > pos:  # 标签之间的内容
                    tokens.append(self.seg.text(data[pos:start]))
                # Find the extent of this tag, if it's ever closed.
                end, found_close = self._tag_extent(data, start)
                pos = end
                if found_close:
                    tag = data[start:end]
                    valid, tag_name, closer, opts = self._parse_tag(tag)
                    # Make sure this is a well-formed, recognized tag, otherwise it's just data.
                    if valid and tag_name in self.tags:
                        if closer:
                            if len(tags[tag_name]) == 0:  # 未找到对应的开始标签
                                tokens.append(self.seg.text(tag))
                                continue
                            # 从右到左, 找到第一个匹配的未闭合的标签, 找到后将其闭合
                            # 期间的表判断 1. tags为空 则加入其中
                            # 2. 有tag, 判断relation, 如果不相同的relation/重复的BBCode 则把其中tag变成text, 并加入tag.
                            #   相同则直接加入tag
                            tag_start = tags[tag_name].pop()
                            bbcode_start = tag_start.tags[0]
                            for i in tokens[::-1]:
                                if isinstance(i, self.seg):  # 是纯文本
                                    if len(i.tags) == 0:
                                        i.extend(tag_start)
                                    else:
                                        if bbcode_start in i.tags or any(
                                            _.relation != bbcode_start.relation for _ in i.tags
                                        ):
                                            for _ in i.tags[::-1]:
                                                i.down(_)
                                        i.extend(tag_start)
                                else:  # 是标签
                                    _tag_name = i[1]
                                    if _tag_name == tag_name:
                                        tokens.remove(i)
                                        break
                        else:
                            seg = self.seg.build("", [self.tags[tag_name]], opts)
                            tags[tag_name].append(seg)
                            tokens.append((seg, tag_name, tag))
                    elif valid and self.drop_unrecognized and tag_name not in self.tags:
                        # If we found a valid (but unrecognized) tag and self.drop_unrecognized is True, just drop it.
                        pass
                    else:
                        tokens.append(self.seg.text(tag))
                else:
                    # We didn't find a closing tag, tack it on as text.
                    tokens.append(self.seg.text(data[start:end]))

            else:  # 没找到标签开始
                break
        if pos < ld:  # 剩余文本
            tokens.append(self.seg.text(data[pos:]))
        for idx, t in enumerate(tokens):
            if not isinstance(t, self.seg):  # 未闭合的标签
                tokens[idx] = self.seg.text(t[2])
        return self.seg.get_message_class()(tokens)



from .adapter import Adapter as Adapter
from .bot import Bot as Bot
from .event import Event as Event
from .event import register_event_class as register_event_class
from .message import Message as Message
from .message import MessageSegment as MessageSegment



from nonebot.internal.permission import Permission

from nonebot_adapter_tailchat.event import DefaultMessageEvent


async def _private(event: DefaultMessageEvent) -> bool:
    return not event.is_group()


async def _group(event: DefaultMessageEvent) -> bool:
    return event.is_group()


GROUP: Permission = Permission(_group)

PRIVATE: Permission = Permission(_private)
